'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Const SmartShader:Int								=	1

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Type TShader

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	GLOBALS
'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Global	program_code:Byte Ptr		
	Global	program_size:Int
	
	Global	AutoFlags:Int						=	0
	
'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	FIELDS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Field		program:Int
	Field		vertexshader:Int
	Field		fragmentshader:Int

	Field		uniform_texture_map:TMap 				=	CreateMap ( )
	Field		uniform_variable_map:TMap 				=	CreateMap ( )
	Field		uniform_map:TMap 						=	CreateMap ( )

	Field		Flags:Int

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	SMARTSHADER FIELDS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Field		Smart_Texture:TShaderUniform [ ]
	Field		Smart_UseTexture:TShaderUniform [ ]

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Update ( )
	
			If	Flags & SmartShader
			
				SmartUpdate_Textures ( )
			
			EndIf
	
			For	Local n:TShaderUniform = EachIn MapValues ( uniform_map )
			
				n.Update ( )

			Next	
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Load ( VertexShader_Url:Object, FragmentShader_Url:Object )
			
			StopDraw ( )

			program							=	glCreateProgramObjectARB ( )
			CheckGLError ( )
			
			vertexshader						=	glCreateShaderObjectARB ( GL_VERTEX_SHADER_ARB )
			CheckGLError ( )

			fragmentshader						=	glCreateShaderObjectARB ( GL_FRAGMENT_SHADER_ARB )	
			CheckGLError ( )
									
			?debug
				
				DebugLog ( "" )
				DebugLog ( "	[ COMPILE VERTEX SHADER ]  " + VertexShader_Url.ToString ( ) )
			
			?

			Upload ( Program, VertexShader, Parse_File ( VertexShader_Url ) )

			?debug
				
				DebugLog ( "	[ COMPILE FRAGMENT SHADER ]  " + FragmentShader_Url.ToString ( ) )
			
			?
			
			Upload ( Program, FragmentShader, Parse_File ( FragmentShader_Url ) )

			For	Local n:TShaderUniform = EachIn MapValues ( uniform_map )
			
				n.uniform						=	glGetUniformLocation ( program, n.Name )
				
				CheckGLError ( )

			Next

			?debug
				
				DebugLog ( "" )
			
			?

	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Parse_File:String ( Url:Object )
	
	Local		f:TStream							=	ReadFile ( Url )
	Local		read:String							=	""
	Local		result:String						=	""
	Local		_rem:Byte							=	False
	
			While Not Eof ( f )
			
				read							=	Trim ( ReadLine ( f ) )
				
				Select True
				
				Case	_rem 
				
					If	Upper ( Left ( read, 6 ) ) = "ENDREM"
				
						_rem					=	False
				
					EndIf

					Continue
				
				Case	read = "", Left ( read, 1 ) = "'", Left ( read, 2 ) = "//"
				
					Continue

				Case	Left ( read, 2 ) = "->"
				
					result					:+	Parse_File ( Trim ( Replace ( read, "->", "" ) ) )

					Continue
				
				Case	Upper ( Left ( read, 3 ) ) = "REM"
				
					_rem						=	True

					Continue
				
				Case	Upper ( Left ( read, 8 ) ) = "UNIFORM "
				
					Parse_Uniform ( read )
				
				EndSelect
			
				result						:+	read + "~n"
				
			Wend	
	
			CloseFile ( f )
	
	Return	( result )
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Parse_Uniform ( source:String )
			
			source							=	Replace ( source, "~t", " " )
			
	Local		Token:String [ ]						=	SplitString ( source, 32 )
	Local		Name:String							=	Upper ( Replace ( Token [ 2 ], ";", "" ) )
	Local		Obj:TShaderUniform					=	New TShaderUniform
	
			Obj.Name							=	Replace ( Token [ 2 ], ";", "" )
			Obj.Shader							=	Self
			
			If	Upper ( Token [ 1 ] ) = "SAMPLER2D"
			
				MapInsert ( uniform_texture_map, Name, Obj )
				MapInsert ( uniform_map, Name, Obj )
				
			Else
			
				MapInsert ( uniform_variable_map, Name, Obj )
				MapInsert ( uniform_map, Name, Obj )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Delete ( )
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	FUNCTIONS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Function	create_program_code ( code:String )
			
			program_code						=	code.ToCString ( )
			program_size						=	code.length
			
	EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Function	upload:Byte ( program:Int, shader:Int, code:String )
			
	Local 	result:Int
			
			create_program_code ( code )
			CheckGLError ( )
				
			glShaderSourceARB ( shader, 1, Varptr program_code, Varptr program_size )
			CheckGLError ( )

			glCompileShaderARB ( shader )
			CheckGLError ( )

			MemFree ( program_code )
			
			glAttachObjectARB ( program, shader )
			CheckGLError ( )

			glDeleteObjectARB ( shader ) 
			CheckGLError ( )

			glGetObjectParameterivARB ( shader, GL_OBJECT_COMPILE_STATUS_ARB, Varptr result )
			CheckGLError ( )


			If	result = False
			
				Local	errlen:Int
				glGetObjectParameterivARB ( shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, Varptr errlen )	
				
				Local	errlog:Byte [ errlen ]						
				glGetInfoLogARB ( shader, errlen + 1, Varptr errlen, Varptr errlog [ 0 ] )

				Error	( "TShader : Shader konnte nicht compiliert werden~n~n" + String.FromCString ( Varptr errlog [ 0 ] ) )

				Return ( False )
	
			EndIf
			
			glLinkProgramARB ( program )
			glGetObjectParameterivARB ( program, GL_OBJECT_LINK_STATUS_ARB, Varptr result )

			If	result = False
			
				Local	errlen:Int
				glGetObjectParameterivARB ( shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, Varptr errlen )	
				
				Local	errlog:Byte [ errlen ]						
				glGetInfoLogARB ( shader, errlen + 1, Varptr errlen, Varptr errlog [ 0 ] )
				
				Error	( "TShader : Programm konnte nicht verlinkt werden~n~n" + String.FromCString ( Varptr errlog [ 0 ] ) )
				
				Return ( False )
				
			EndIf
			
			?debug

				Local	errlen:Int
				glGetObjectParameterivARB ( shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, Varptr errlen )	
				
				Local	errlog:Byte [ errlen ]						
				glGetInfoLogARB ( shader, errlen + 1, Varptr errlen, Varptr errlog [ 0 ] )
	
				DebugLog ( "		" + String.FromCString ( Varptr errlog [ 0 ] ) )
				
			?
			
			Return ( True )
								
	EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	SmartUpdate_Textures ( )
	
	Local		MaxUnits:Int						=	MaxTextureUnits ( )
			
			If	Smart_Texture = Null
				
				Smart_Texture					=	New TShaderUniform [ MaxUnits ]
				Smart_UseTexture					=	New TShaderUniform [ MaxUnits ]
			
				For	Local Unit:Int = 0 Until MaxUnits
		
					Smart_Texture [ Unit ]			=	GetUniform ( Self, "Texture_" + Unit )
					Smart_UseTexture [ Unit ]		=	GetUniform ( Self, "UseTexture_" + Unit )

					If	Smart_Texture [ Unit ]
						
						UpdateUniform ( Smart_Texture [ Unit ], [ Unit ] )
					
					EndIf
	
					If	Smart_UseTexture [ Unit ]
					
						UpdateUniform ( Smart_UseTexture [ Unit ], [ 0 ] )
					
					EndIf

				Next
			
			EndIf
			
			For	Local Unit:Int = 0 Until MaxUnits
	
				If	Smart_UseTexture [ Unit ]
				
					UpdateUniform ( Smart_UseTexture [ Unit ], [ ( GetTexture ( Unit ) <> Null ) * _AVA_GFX_TEXTURE_SMARTSHADER_MODE [ Unit ] . Get ( ) ] )
				
				EndIf
			
			Next
			
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	ENDTYPE
'---------------------------------------------------------------------------------------------------------------------------------------------------------

EndType

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Type TShaderUniform

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Field name:String								=	""
	Field uniform:Int								=	-1
	Field	shader:TShader							=	Null
	Field	vi:Int [ ]								=	Null
	Field	vf:Float [ ]							=	Null
	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	Update ( )
			
			If	_AVA_GFX_ACTIVE_SHADER	= Shader

				StopDraw ( )
			
				Select True
				
				Case	vi.length = 1					Set1i ( vi [ 0 ] )	
				Case	vi.length = 2					Set2i ( vi [ 0 ], vi [ 1 ] )	
				Case	vi.length = 3					Set3i ( vi [ 0 ], vi [ 1 ], vi [ 2 ] )	
				Case	vi.length => 4					Set4i ( vi [ 0 ], vi [ 1 ], vi [ 2 ], vi [ 3 ] )	
				
				Case	vf.length = 1					Set1f ( vf [ 0 ] )	
				Case	vf.length = 2					Set2f ( vf [ 0 ], vf [ 1 ] )	
				Case	vf.length = 3					Set3f ( vf [ 0 ], vf [ 1 ], vf [ 2 ] )	
				Case	vf.length => 4					Set4f ( vf [ 0 ], vf [ 1 ], vf [ 2 ], vf [ 3 ] )			
				
				EndSelect
				
				vi							=	Null
				vf							=	Null
			
			EndIf
			
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set1i ( a:Int )
	
			If	uniform > -1
			
				glUniform1i ( uniform, a )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set2i ( a:Int, b:Int )
	
			If	uniform > -1
			
				glUniform2i ( uniform, a, b )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set3i ( a:Int, b:Int, c:Int )
	
			If	uniform > -1
			
				glUniform3i ( uniform, a, b, c )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set4i ( a:Int, b:Int, c:Int, d:Int )
	
			If	uniform > -1
			
				glUniform4i ( uniform, a, b, c, d )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set1f ( a:Float )
	
			If	uniform > -1
			
				glUniform1f ( uniform, a )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set2f ( a:Float, b:Float )
	
			If	uniform > -1
			
				glUniform2f ( uniform, a, b )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set3f ( a:Float, b:Float, c:Float )
	
			If	uniform > -1
			
				glUniform3f ( uniform, a, b, c )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------
	
	Method	Set4f ( a:Float, b:Float, c:Float, d:Float )
	
			If	uniform > -1
			
				glUniform4f ( uniform, a, b, c, d )
			
			EndIf
	
	EndMethod

'---------------------------------------------------------------------------------------------------------------------------------------------------------

EndType

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

