'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	THIGHSCORE
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Type THighscore

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	FIELDS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Field		entry_list:TList						=	CreateList()
	Field		max_entries:Int, password:String

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	METHODS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	add ( n:THighscoreEntry )
	
			ListAddLast ( entry_list, n )
			
			update
	
	EndMethod
	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	update ( )
	
			SortList ( entry_list )
			
			Local	i:Int							=	0
				
			For	Local n:THighscoreEntry = EachIn entry_list
				
				i							:+	1
				
				If	i > max_entries
			
					n.position					=	0
					ListRemove ( entry_list, n )
				
				Else
				
					n.position					=	i
				
				EndIf
			
			Next
			
	EndMethod
	
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	to_stream ( f:TStream )
	
	Local		p:Int								=	0
			
			WriteInt ( f, max_entries )
			WriteInt ( f, CountList(entry_list) )
			
			For	Local n:THighscoreEntry = EachIn entry_list

				WriteInt ( f, n.position )
				WriteInt ( f, n.name.length )
				WriteString ( f, decode(n.name,p) )
				WriteLong ( f, n.score )
				WriteInt ( f, n.infos.length )
				
				p							:+	n.score
				
				For	Local i:Int = 0 Until n.infos.length
					
					WriteInt ( f, n.infos[i].length )
					WriteString ( f, decode(n.infos[i],p) )				
				
				Next

			Next

			WriteString ( f, decode("highscore_checksum_okay!",p) )				
						
	EndMethod
		
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	from_stream:String ( f:TStream )
	
	Local		p:Int								=	0
	Local		num:Int							=	0
	Local		length:Int							=	0
	
			max_entries							=	ReadInt ( f )
			num								=	ReadInt ( f )
			
			For	Local i:Int = 0 Until num
				
				If	Eof ( f )
					
					Return ( "" )
				
				EndIf
					
				Local	n:THighscoreEntry				=	New THighscoreEntry
				
				n.position						=	ReadInt ( f )
				length						=	ReadInt ( f )
				
				If	length > 256
				
					Return ( "" )
				
				EndIf
				
				n.name						=	decode( ReadString ( f, length ), p )
				n.score						=	ReadLong ( f )
				n.infos						=	n.infos [ .. ReadInt ( f ) ]

				p							:+	n.score
				
				For	Local i:Int = 0 Until n.infos.length
					
					length					=	ReadInt ( f )
					
					If	length > 256
					
						Return ( "" )
					
					EndIf
					
					n.infos [ i ]				=	decode ( ReadString ( f, length ), p )			
				
				Next
				
				add ( n )

			Next
	
			update
			
	Return	decode( ReadString ( f, 24 ), p )		
						
	EndMethod
		
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	decode:String ( n:String, p:Int Var  )
	
	Local		result:String						=	""

			For	Local i:Int = 0 Until n.length

				p							:+	1				
				result						=	result + Chr( n[i] ~ password[p Mod password.length] )

			Next
			
	Return	result
						
	EndMethod
		
'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	ENDTYPE
'---------------------------------------------------------------------------------------------------------------------------------------------------------

EndType

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	THIGHSCOREENTRY
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc: NA
End Rem
Type THighscoreEntry

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	FIELDS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Field		position:Int, name:String, score:Long, infos:String[]

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	METHODS
'---------------------------------------------------------------------------------------------------------------------------------------------------------

	Method	compare:Int ( obj:Object )
	
	Local		n:THighscoreEntry						=	THighscoreEntry ( obj )
	
			If	n
			
				If	n = Self
				
					Return ( 0 )
				
				EndIf
			
				Select ( n.score > score )					
				Case	True	Return ( 1 )
				Case	False	Return ( -1 )
				EndSelect
				
			EndIf
	
	EndMethod
	
'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	ENDTYPE
'---------------------------------------------------------------------------------------------------------------------------------------------------------

EndType

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	HIGHSCORE
'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function CreateHighscore:THighscore ( password:String, max_entries:Int = 10 )

Local		n:THighscore							=	New THighscore

		n.password								=	password
		n.max_entries							=	max_entries

Return	n

EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function AddHighscore:THighscoreEntry ( highscore:THighscore, name:String, score:Long, infos:String[] = Null )

If		highscore = Null								Error (  "AddHighscore : THighscore ist NULL." )
Local		n:THighscoreEntry 						=	New THighscoreEntry 

		n.name								=	name
		n.score								=	score
		n.infos								=	infos

		highscore.add ( n )

Return	n

EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function CheckHighscore:Int ( highscore:THighscore, score:Long )

If		highscore = Null								Error (  "CheckHighscore : THighscore ist NULL." )

		For	Local n:THighscoreEntry = EachIn highscore.entry_list
			
			If	score > n.score

				Return ( n.position )
	
			EndIf
				
		Next

Return	0

EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function ListHighscore ( highscore:THighscore, func(position:Int,name:String,score:Long,infos:String[]) )

If		highscore = Null								Error ( "ListHighscore : THighscore ist NULL." )
If		func = Null									Error ( "ListHighscore : Funktionspointer ist NULL." )

		For	Local n:THighscoreEntry = EachIn highscore.entry_list
			
			func ( n.position, n.name, n.score, n.infos )
				
		Next

EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function LoadHighscore:THighscore ( url:Object, password:String )

Local		stream:TStream							=	ReadStream ( url )
If		stream = Null								Error ( "LoadHighscore : ~q" + url.tostring() +"~q konnte nicht gelesen werden." )
Local		n:THighscore							=	New THighscore
		
		n.password								=	password
		
If		n.from_stream ( stream ) <> "highscore_checksum_okay!"	Error (  "LoadHighscore : Die Datei ~q" + url.tostring() + "~q  wurde manipuliert." )	
	
		CloseStream ( stream )

Return	n

EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------

Rem
bbdoc:
End Rem
Function SaveHighscore ( highscore:THighscore, url:Object )

If		highscore = Null								Error ( "SaveHighscore : THighscore ist NULL." )
Local		stream:TStream							=	WriteStream ( url )
If		stream = Null								Error ( "SaveHighscore : ~q" + url.tostring() +"~q konnte nicht gespeichert werdenn." )

		highscore.to_stream ( stream )
		
		CloseStream ( stream )
		
EndFunction

'---------------------------------------------------------------------------------------------------------------------------------------------------------
'	FINISHED
'---------------------------------------------------------------------------------------------------------------------------------------------------------
